#!/usr/bin/env python3
"""
Test the UUID class in `btle.py`

Run with:
    $ python -m unittest this_file.py
"""

import binascii
import unittest

from bluepy3.btle import UUID


def cmp(a, b):
    return (a > b) - (a < b)


class TestUUID(unittest.TestCase):
    def test_init_with_good_uuids(self):
        self.assertIsInstance(UUID(10880), UUID)
        self.assertIsInstance(UUID("10880"), UUID)
        self.assertIsInstance(UUID(10007), UUID)
        self.assertIsInstance(UUID("10007"), UUID)
        self.assertIsInstance(UUID("000a0200-0000-1000-7000-00609bf534fb"), UUID)
        self.assertIsInstance(UUID("000a020000001000700000609bf534fb"), UUID)
        self.assertIsInstance(UUID("12345678-5001-abcd-3422-dbc9eac7543c"), UUID)
        self.assertIsInstance(UUID("123456785001abcd3422dbc9eac7543c"), UUID)
        self.assertIsInstance(UUID("99999999-9999-9999-9999-999999999999"), UUID)
        self.assertIsInstance(UUID("99999999999999999999999999999999"), UUID)
        self.assertIsInstance(UUID("00000000-0000-0000-0000-000000000000"), UUID)
        self.assertIsInstance(UUID("00000000000000000000000000000000"), UUID)
        self.assertIsInstance(UUID(0), UUID)
        self.assertIsInstance(UUID("0"), UUID)
        self.assertIsInstance(UUID(0x0), UUID)
        self.assertIsInstance(UUID(0000), UUID)
        self.assertIsInstance(UUID("0000"), UUID)
        self.assertIsInstance(UUID(0x0000), UUID)
        self.assertIsInstance(UUID(00000000), UUID)
        self.assertIsInstance(UUID("00000000"), UUID)
        self.assertIsInstance(UUID(0x00000000), UUID)
        self.assertIsInstance(UUID("F"), UUID)
        self.assertIsInstance(UUID("f"), UUID)
        self.assertIsInstance(UUID(0xF), UUID)
        self.assertIsInstance(UUID("FFFF"), UUID)
        self.assertIsInstance(UUID(0xFFFF), UUID)
        self.assertIsInstance(UUID("FFFFFFFF"), UUID)
        self.assertIsInstance(UUID(0xFFFFFFFF), UUID)

    def test_init_with_bad_uuids(self):
        self.assertRaises(ValueError, UUID, -12345)
        self.assertRaises(ValueError, UUID, -1)
        self.assertRaises(ValueError, UUID, "ABCDEFG")
        self.assertRaises(ValueError, UUID, ":::::::::")
        self.assertRaises(ValueError, UUID, '!@#?><:"567')
        self.assertRaises(ValueError, UUID, "12345678-4912-e123-3333-abcdef1234")
        self.assertRaises(ValueError, UUID, "123456784912e1233333abcdef1234")
        self.assertRaises(ValueError, UUID, "This is a sentence.")
        self.assertRaises(ValueError, UUID, 0xFFFFFFFFA)
        self.assertRaises(ValueError, UUID, 0xFFFFFFFFBB)
        self.assertRaises(binascii.Error, UUID, "0x0")
        self.assertRaises(binascii.Error, UUID, "0xFFFFFFFFA")
        self.assertRaises(binascii.Error, UUID, "0xFFFFFFFFBB")

    def test_str(self):
        self.assertEqual(UUID(0), "00000000-0000-1000-8000-00805f9b34fb")
        self.assertEqual(UUID("0"), "00000000-0000-1000-8000-00805f9b34fb")
        self.assertEqual(UUID(0x0), "00000000-0000-1000-8000-00805f9b34fb")
        self.assertEqual(UUID(0000), "00000000-0000-1000-8000-00805f9b34fb")
        self.assertEqual(UUID("0000"), "00000000-0000-1000-8000-00805f9b34fb")
        self.assertEqual(UUID(0x0000), "00000000-0000-1000-8000-00805f9b34fb")
        self.assertEqual(UUID("F"), "0000000f-0000-1000-8000-00805f9b34fb")
        self.assertEqual(UUID("f"), "0000000f-0000-1000-8000-00805f9b34fb")
        self.assertEqual(UUID("FFFF"), "0000ffff-0000-1000-8000-00805f9b34fb")
        self.assertEqual(UUID("ABCD"), "0000abcd-0000-1000-8000-00805f9b34fb")
        self.assertEqual(UUID(0xFFFFFFFF), "ffffffff-0000-1000-8000-00805f9b34fb")
        self.assertEqual(UUID(0xFFFFFFFF), "ffffffff-0000-1000-8000-00805f9b34fb")
        self.assertEqual(
            UUID("00000000-0000-1000-8000-00805f9b34fb"),
            "00000000-0000-1000-8000-00805f9b34fb",
        )
        self.assertEqual(
            UUID("0000000000001000800000805f9b34fb"),
            "00000000-0000-1000-8000-00805f9b34fb",
        )

    def test_eq(self):
        self.assertTrue(UUID(0) == "00000000-0000-1000-8000-00805f9b34fb")
        self.assertTrue(UUID("0") == "00000000-0000-1000-8000-00805f9b34fb")
        self.assertTrue(UUID(0x0) == "00000000-0000-1000-8000-00805f9b34fb")
        self.assertTrue(UUID(0000) == "00000000-0000-1000-8000-00805f9b34fb")
        self.assertTrue(UUID("0000") == "00000000-0000-1000-8000-00805f9b34fb")
        self.assertTrue(UUID("F") == "0000000f-0000-1000-8000-00805f9b34fb")
        self.assertTrue(UUID("f") == "0000000f-0000-1000-8000-00805f9b34fb")
        self.assertTrue(UUID("FFFF") == "0000ffff-0000-1000-8000-00805f9b34fb")
        self.assertTrue(UUID("ABCD") == "0000abcd-0000-1000-8000-00805f9b34fb")
        self.assertTrue(UUID(0xFFFFFFFF) == "ffffffff-0000-1000-8000-00805f9b34fb")
        self.assertTrue(UUID(0xABCDEF12) == "abcdef12-0000-1000-8000-00805f9b34fb")
        self.assertTrue(UUID("00000000-0000-1000-8000-00805f9b34fb") == "00000000-0000-1000-8000-00805f9b34fb")
        self.assertTrue(UUID("0000000000001000800000805f9b34fb") == "00000000-0000-1000-8000-00805f9b34fb")

    def test_cmp(self):
        self.assertTrue(cmp(UUID(0), "00000000-0000-1000-8000-00805f9b34fb"))
        self.assertTrue(cmp(UUID("0"), "00000000-0000-1000-8000-00805f9b34fb"))
        self.assertTrue(cmp(UUID(0x0), "00000000-0000-1000-8000-00805f9b34fb"))
        self.assertTrue(cmp(UUID(0000), "00000000-0000-1000-8000-00805f9b34fb"))
        self.assertTrue(cmp(UUID("0000"), "00000000-0000-1000-8000-00805f9b34fb"))
        self.assertTrue(cmp(UUID("F"), "0000000f-0000-1000-8000-00805f9b34fb"))
        self.assertTrue(cmp(UUID("f"), "0000000f-0000-1000-8000-00805f9b34fb"))
        self.assertTrue(cmp(UUID("FFFF"), "0000ffff-0000-1000-8000-00805f9b34fb"))
        self.assertTrue(cmp(UUID("ABCD"), "0000abcd-0000-1000-8000-00805f9b34fb"))
        self.assertTrue(cmp(UUID(0xFFFFFFFF), "ffffffff-0000-1000-8000-00805f9b34fb"))
        self.assertTrue(cmp(UUID(0xABCDEF12), "abcdef12-0000-1000-8000-00805f9b34fb"))
        self.assertTrue(
            cmp(
                UUID("00000000-0000-1000-8000-00805f9b34fb"),
                "00000000-0000-1000-8000-00805f9b34fb",
            )
        )
        self.assertTrue(
            cmp(
                UUID("0000000000001000800000805f9b34fb"),
                "00000000-0000-1000-8000-00805f9b34fb",
            )
        )


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3

import argparse
import binascii
import os
import sys

from . import btle

if os.getenv("C", "1") == "0":
    ANSI_RED = ""
    ANSI_GREEN = ""
    ANSI_YELLOW = ""
    ANSI_CYAN = ""
    ANSI_WHITE = ""
    ANSI_OFF = ""
else:
    ANSI_CSI = "\033["
    ANSI_RED = ANSI_CSI + "31m"
    ANSI_GREEN = ANSI_CSI + "32m"
    ANSI_YELLOW = ANSI_CSI + "33m"
    ANSI_CYAN = ANSI_CSI + "36m"
    ANSI_WHITE = ANSI_CSI + "37m"
    ANSI_OFF = ANSI_CSI + "0m"


def dump_services(dev):
    services = sorted(dev.services, key=lambda k: k.hndStart)
    for s in services:
        print(f"\t{s.hndStart:04X}: {s}")
        if s.hndStart == s.hndEnd:
            continue
        chars = s.getCharacteristics()
        for i, c in enumerate(chars):
            props = c.propertiesToString()
            h = c.getHandle()
            if "READ" in props:
                val = c.read()
                if c.uuid == btle.AssignedNumbers.device_name:
                    string = ANSI_CYAN + "'" + val.decode("utf-8") + "'" + ANSI_OFF
                elif c.uuid == btle.AssignedNumbers.device_information:
                    string = repr(val)
                else:
                    string = "<s" + binascii.b2a_hex(val).decode("utf-8") + ">"
            else:
                string = ""
            print(f"\t{h:04X}:    {c} {props:>59} {string:>12}")

            while True:
                h += 1
                if h > s.hndEnd or (i < len(chars) - 1 and h >= chars[i + 1].getHandle() - 1):
                    break
                try:
                    val = dev.readCharacteristic(h)
                    bval = binascii.b2a_hex(val).decode("utf-8")
                    print(f"\t{h:04x}:     <{bval}>")
                except btle.BTLEException:
                    break


class ScanPrint(btle.DefaultDelegate):
    def __init__(self, opts):
        btle.DefaultDelegate.__init__(self)
        self.opts = opts

    def handleDiscovery(self, dev, isNewDev, isNewData):
        if isNewDev:
            status = "new"
        elif isNewData:
            if self.opts.new:
                return
            status = "update"
        else:
            if not self.opts.all:
                return
            status = "old"

        if dev.rssi < self.opts.sensitivity:
            return
        dev_connectable = "(not connectable)"
        if dev.connectable:
            dev_connectable = ""
        print(
            f"    Device ({status}): {ANSI_WHITE}{dev.addr}{ANSI_OFF} ({dev.addrType}),"
            f" {dev.rssi} dBm {dev_connectable}"
        )
        for (sdid, desc, val) in dev.getScanData():
            if sdid in [8, 9]:
                print(f"\t{desc}: '{ANSI_CYAN}{val}{ANSI_OFF}'")
            else:
                print(f"\t{desc}: <{val}>")
        if not dev.scanData:
            print("\t(no data)")
        print()


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-i",
        "--hci",
        action="store",
        type=int,
        default=0,
        help="Interface number for scan",
    )
    parser.add_argument(
        "-t",
        "--timeout",
        action="store",
        type=int,
        default=4,
        help="Scan delay, 0 for continuous",
    )
    parser.add_argument(
        "-s",
        "--sensitivity",
        action="store",
        type=int,
        default=-128,
        help="dBm value for filtering far devices",
    )
    parser.add_argument(
        "-d",
        "--discover",
        action="store_true",
        help="Connect and discover service to scanned devices",
    )
    parser.add_argument(
        "-a",
        "--all",
        action="store_true",
        help="Display duplicate adv responses, by default show new + updated",
    )
    parser.add_argument(
        "-n",
        "--new",
        action="store_true",
        help="Display only new adv responses, by default show new + updated",
    )
    parser.add_argument("-v", "--verbose", action="store_true", help="Increase output verbosity")
    arg = parser.parse_args(sys.argv[1:])

    btle.Debugging = arg.verbose

    scanner = btle.Scanner(arg.hci).withDelegate(ScanPrint(arg))

    print(ANSI_RED + "Scanning for devices..." + ANSI_OFF)
    devices = scanner.scan(arg.timeout)

    if arg.discover:
        print(ANSI_RED + "Discovering services..." + ANSI_OFF)

        for d in devices:
            if not d.connectable or d.rssi < arg.sensitivity:
                continue

            print("    Connecting to", ANSI_WHITE + d.addr + ANSI_OFF + ":")

            dev = btle.Peripheral(d)
            dump_services(dev)
            dev.disconnect()
            print()


if __name__ == "__main__":
    main()



"""Python setup script for bluepy3"""

import os
import shlex
import subprocess
import sys

from setuptools import setup
from setuptools.command.build_py import build_py

VERSION = "0.0.25"


def pre_install():
    """Do the custom compiling of the bluepy3-helper executable from the makefile"""
    cmd = ""
    try:
        print("\n\n*** Executing pre-install ***\n")
        print(f"Working dir is {os.getcwd()}")
        with open("bluepy3/version.h", "w") as verfile:
            verfile.write(f'#define VERSION_STRING "{VERSION}"\n')
        for cmd in ["make -dC bluepy3 clean", "make -dC bluepy3 -j1"]:
            print(f"\nexecute {cmd}")
            msgs = subprocess.check_output(shlex.split(cmd), stderr=subprocess.STDOUT)  # noqa
        print("\n\n*** Finished pre-install ***\n\n")
    except subprocess.CalledProcessError as e:
        print("Failed to compile bluepy3-helper. Exiting install.")
        print(f"Command was {repr(cmd)} in {os.getcwd()}")
        print(f"Return code was {e.returncode}")
        print(f"Output was:\n{e.output}")
        sys.exit(1)


class MyBuildPy(build_py):
    def run(self):
        pre_install()
        build_py.run(self)


setup_cmdclass = {
    "build_py": MyBuildPy,
}

# Force package to be *not* pure Python
# Discusssed at issue #158

try:
    from wheel.bdist_wheel import bdist_wheel  # noqa

    class BluepyBdistWheel(bdist_wheel):
        def finalize_options(self):
            bdist_wheel.finalize_options(self)
            self.root_is_pure = False  # noqa

    setup_cmdclass["bdist_wheel"] = BluepyBdistWheel
except ImportError:
    pass

setup(
    name="bluepy3",
    version=VERSION,
    description="Python module for interfacing with BLE devices through Bluez",
    author="mausy5043",
    url="https://github.com/Mausy5043/bluepy3",
    download_url="https://github.com/Mausy5043/bluepy3",
    keywords=["Bluetooth", "Bluetooth Smart", "BLE", "Bluetooth Low Energy"],
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    packages=["bluepy3"],
    package_data={
        "bluepy3": [
            "bluepy3-helper",
            "*.json",
            "bluepy3-helper.c",
            "config.*.h",
            "version.h",
            "Makefile",
        ]
    },
    cmdclass=setup_cmdclass,
    entry_points={
        "console_scripts": [
            "thingy52=bluepy3.thingy52:main",
            "sensortag=bluepy3.sensortag:main",
            "blescan=bluepy3.blescan:main",
        ]
    },
)



#!/usr/bin/env python3

"""Bluetooth Low Energy Python3 interface"""

import sys
import os
import time
import subprocess
import binascii
import struct
import signal
from queue import Queue, Empty
from threading import Thread


def preexec_function():
    # Ignore the SIGINT signal by setting the handler to the standard
    # signal handler SIG_IGN.
    signal.signal(signal.SIGINT, signal.SIG_IGN)


Debugging = False
script_path = os.path.join(os.path.abspath(os.path.dirname(__file__)))
helperExe = os.path.join(script_path, "bluepy3-helper")

SEC_LEVEL_LOW = "low"
SEC_LEVEL_MEDIUM = "medium"
SEC_LEVEL_HIGH = "high"

ADDR_TYPE_PUBLIC = "public"
ADDR_TYPE_RANDOM = "random"


def DBG(*args):
    if Debugging:
        msg = " ".join([str(a) for a in args])
        print(msg)


class BTLEException(Exception):
    """Base class for all Bluepy exceptions"""

    def __init__(self, message, resp_dict=None):
        self.message = message

        # optional messages from bluepy3-helper
        self.estat = None
        self.emsg = None
        if resp_dict:
            self.estat = resp_dict.get("estat", None)
            if isinstance(self.estat, list):
                self.estat = self.estat[0]
            self.emsg = resp_dict.get("emsg", None)
            if isinstance(self.emsg, list):
                self.emsg = self.emsg[0]

    def __str__(self):
        msg = self.message
        if self.estat or self.emsg:
            msg = f"{msg} ("
            if self.estat:
                msg = f"{msg}code: {self.estat}"
            if self.estat and self.emsg:
                msg = f"{msg}, "
            if self.emsg:
                msg = f"{msg}error: {self.emsg}"
            msg = f"{msg})"
        return msg


class BTLEInternalError(BTLEException):
    def __init__(self, message, rsp=None):
        BTLEException.__init__(self, message, rsp)


class BTLEDisconnectError(BTLEException):
    def __init__(self, message, rsp=None):
        BTLEException.__init__(self, message, rsp)


class BTLEManagementError(BTLEException):
    def __init__(self, message, rsp=None):
        BTLEException.__init__(self, message, rsp)


class BTLEGattError(BTLEException):
    def __init__(self, message, rsp=None):
        BTLEException.__init__(self, message, rsp)


class UUID:
    def __init__(self, val, commonName=None):
        """Initialisation
        We accept: 32-digit hex strings, with and without '-' characters,
        4 to 8 digit hex strings, and integers
        """
        if isinstance(val, int):
            if (val < 0) or (val > 0xFFFFFFFF):
                raise ValueError("Short form UUIDs must be in range 0..0xFFFFFFFF")
            val = f"{val:04X}"
        elif isinstance(val, self.__class__):
            val = str(val)
        else:
            val = str(val)  # Do our best

        val = val.replace("-", "")
        if len(val) <= 8:  # Short form
            val = ("0" * (8 - len(val))) + val + "00001000800000805F9B34FB"

        self.binVal = binascii.a2b_hex(val.encode("utf-8"))
        if len(self.binVal) != 16:
            raise ValueError(f"UUID must be 16 bytes, got '{val}' (len={len(self.binVal)})")
        self.commonName = commonName

    def __str__(self):
        s = binascii.b2a_hex(self.binVal).decode("utf-8")
        return "-".join([s[0:8], s[8:12], s[12:16], s[16:20], s[20:32]])

    def __eq__(self, other):
        return self.binVal == UUID(other).binVal

    # def __cmp__(self, other):
    #     return cmp(self.binVal, UUID(other).binVal)

    def __hash__(self):
        return hash(self.binVal)

    def getCommonName(self):
        s = AssignedNumbers.getCommonName(self)
        if s:
            return s
        s = str(self)
        if s.endswith("-0000-1000-8000-00805f9b34fb"):
            s = s[0:8]
            if s.startswith("0000"):
                s = s[4:]
        return s


class Service:
    def __init__(self, *args):
        (self.peripheral, uuidVal, self.hndStart, self.hndEnd) = args
        self.uuid = UUID(uuidVal)
        self.chars = None
        self.descs = None

    def getCharacteristics(self, forUUID=None):
        if not self.chars:  # Unset, or empty
            self.chars = (
                [] if self.hndEnd <= self.hndStart else self.peripheral.getCharacteristics(self.hndStart, self.hndEnd)
            )
        if forUUID is not None:
            u = UUID(forUUID)
            return [ch for ch in self.chars if ch.uuid == u]
        return self.chars

    def getDescriptors(self, forUUID=None):
        if not self.descs:
            # Grab all descriptors in our range, except for the service
            # declaration descriptor
            all_descs = self.peripheral.getDescriptors(self.hndStart + 1, self.hndEnd)
            # Filter out the descriptors for the characteristic properties
            # Note that this does not filter out characteristic value descriptors
            self.descs = [desc for desc in all_descs if desc.uuid != 0x2803]
        if forUUID is not None:
            u = UUID(forUUID)
            return [desc for desc in self.descs if desc.uuid == u]
        return self.descs

    def __str__(self):
        return f"Service <uuid={self.uuid.getCommonName()} handleStart={self.hndStart} handleEnd={self.hndEnd}>"


class Characteristic:
    # Currently only READ is used in supportsRead function,
    # the rest is included to facilitate supportsXXXX functions if required
    props = {
        "BROADCAST": 0b00000001,
        "READ": 0b00000010,
        "WRITE_NO_RESP": 0b00000100,
        "WRITE": 0b00001000,
        "NOTIFY": 0b00010000,
        "INDICATE": 0b00100000,
        "WRITE_SIGNED": 0b01000000,
        "EXTENDED": 0b10000000,
    }

    propNames = {
        0b00000001: "BROADCAST",
        0b00000010: "READ",
        0b00000100: "WRITE NO RESPONSE",
        0b00001000: "WRITE",
        0b00010000: "NOTIFY",
        0b00100000: "INDICATE",
        0b01000000: "WRITE SIGNED",
        0b10000000: "EXTENDED PROPERTIES",
    }

    def __init__(self, *args):
        (self.peripheral, uuidVal, self.handle, self.properties, self.valHandle) = args
        self.uuid = UUID(uuidVal)
        self.descs = None

    def read(self):
        return self.peripheral.readCharacteristic(self.valHandle)

    def write(self, val, withResponse=False):
        return self.peripheral.writeCharacteristic(self.valHandle, val, withResponse)

    def getDescriptors(self, forUUID=None, hndEnd=0xFFFF):
        if not self.descs:
            # Descriptors (not counting the value descriptor) begin after
            # the handle for the value descriptor and stop when we reach
            # the handle for the next characteristic or service
            self.descs = []
            for desc in self.peripheral.getDescriptors(self.valHandle + 1, hndEnd):
                if desc.uuid in (0x2800, 0x2801, 0x2803):
                    # Stop if we reach another characteristic or service
                    break
                self.descs.append(desc)
        if forUUID is not None:
            u = UUID(forUUID)
            return [desc for desc in self.descs if desc.uuid == u]
        return self.descs

    def __str__(self):
        return f"Characteristic <{self.uuid.getCommonName()}>"

    def supportsRead(self):
        if self.properties & Characteristic.props["READ"]:
            return True
        else:
            return False

    def propertiesToString(self):
        propStr = ""
        for p in Characteristic.propNames:
            if p & self.properties:
                propStr += Characteristic.propNames[p] + " "
        return propStr

    def getHandle(self):
        return self.valHandle


class Descriptor:
    def __init__(self, *args):
        (self.peripheral, uuidVal, self.handle) = args
        self.uuid = UUID(uuidVal)

    def __str__(self):
        return f"Descriptor <{self.uuid.getCommonName()}>"

    def read(self):
        return self.peripheral.readCharacteristic(self.handle)

    def write(self, val, withResponse=False):
        self.peripheral.writeCharacteristic(self.handle, val, withResponse)


class DefaultDelegate:
    def __init__(self):
        pass

    def handleNotification(self, cHandle, data):
        DBG("Notification:", cHandle, "sent data", binascii.b2a_hex(data))

    def handleDiscovery(self, scanEntry, isNewDev, isNewData):
        DBG("Discovered device", scanEntry.addr)


class BluepyHelper:
    def __init__(self):
        self._helper = None
        self._lineq = None
        self._stderr = None
        self._mtu = 0
        self.delegate = DefaultDelegate()

    def withDelegate(self, delegate_):
        self.delegate = delegate_
        return self

    def _startHelper(self, iface=None):
        if self._helper is None:
            DBG("Running ", helperExe)
            self._aiti = 0
            self._lineq = Queue()
            self._mtu = 0
            self._stderr = open(os.devnull, "w")
            args = [helperExe]
            if iface is not None:
                args.append(str(iface))
            self._helper = subprocess.Popen(
                args,
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=self._stderr,
                universal_newlines=True,
                preexec_fn=preexec_function,
            )
            t = Thread(target=self._readToQueue)
            t.daemon = True  # don't wait for it to exit
            t.start()

    def _readToQueue(self):
        """Thread to read lines from stdout and insert in queue."""
        while self._helper:
            line = self._helper.stdout.readline()
            if not line:  # EOF
                break
            self._lineq.put(line)

    def _stopHelper(self):
        if self._helper is not None:
            DBG("Stopping ", helperExe)
            self._helper.stdin.write("quit\n")
            self._helper.stdin.flush()
            self._helper.wait()
            self._helper = None
            self._aiti = None
        if self._stderr is not None:
            self._stderr.close()
            self._stderr = None

    def _writeCmd(self, cmd):
        if self._helper is None:
            raise BTLEInternalError("Helper not started (did you call connect()?)")
        DBG("Sent: ", cmd)
        self._helper.stdin.write(cmd)
        self._helper.stdin.flush()

    def _mgmtCmd(self, cmd):
        self._writeCmd(cmd + "\n")
        rsp = self._waitResp("mgmt")
        if rsp["code"][0] != "success":
            self._stopHelper()
            raise BTLEManagementError(f"Failed to execute management command '{cmd}'", rsp)

    @staticmethod
    def parseResp(line):
        resp = {}
        for item in line.rstrip().split("\x1e"):
            (tag, tval) = item.split("=")
            if len(tval) == 0:
                val = None
            elif tval[0] == "$" or tval[0] == "'":
                # Both symbols and strings as Python strings
                val = tval[1:]
            elif tval[0] == "h":
                val = int(tval[1:], 16)
            elif tval[0] == "b":
                val = binascii.a2b_hex(tval[1:].encode("utf-8"))
            else:
                raise BTLEInternalError(f"Cannot understand response value {repr(tval)}")
            if tag not in resp:
                resp[tag] = [val]
            else:
                resp[tag].append(val)
        return resp

    def _waitResp(self, wantType, timeout=None):
        while True:
            if self._helper.poll() is not None:
                raise BTLEInternalError("Helper exited")

            try:
                rv = self._lineq.get(timeout=timeout)
            except Empty:
                DBG("Select timeout")
                return None

            DBG("Got:", repr(rv))
            if rv.startswith("#") or rv == "\n" or len(rv) == 0:
                continue

            resp = BluepyHelper.parseResp(rv)
            if "rsp" not in resp:
                raise BTLEInternalError("No response type indicator", resp)

            # sometimes devices just keep sending `ntfy`
            if "ntfy" in repr(rv):
                self._aiti += 1
                if self._aiti > 3:
                    self._stopHelper()
                    raise BTLEInternalError("I am not an idiot.", resp)

            respType = resp["rsp"][0]

            # always check for MTU updates
            if "mtu" in resp and len(resp["mtu"]) > 0:
                new_mtu = int(resp["mtu"][0])
                if self._mtu != new_mtu:
                    self._mtu = new_mtu
                    DBG("Updated MTU: " + str(self._mtu))

            if respType in wantType:
                return resp
            elif respType == "stat":
                if "state" in resp and len(resp["state"]) > 0 and resp["state"][0] == "disc":
                    self._stopHelper()
                    raise BTLEDisconnectError("Device disconnected", resp)
            elif respType == "err":
                errcode = resp["code"][0]
                if errcode == "nomgmt":
                    raise BTLEManagementError("Management not available (permissions problem?)", resp)
                elif errcode == "atterr":
                    raise BTLEGattError("Bluetooth command failed", resp)
                else:
                    raise BTLEException(f"Error from bluepy3-helper ({errcode})", resp)
            elif respType == "scan":
                # Scan response when we weren't interested. Ignore it
                continue
            else:
                raise BTLEInternalError(f"Unexpected response ({respType})", resp)

    def status(self):
        self._writeCmd("stat\n")
        return self._waitResp(["stat"])


class Peripheral(BluepyHelper):
    def __init__(self, deviceAddr=None, addrType=ADDR_TYPE_PUBLIC, iface=None, timeout=None):
        BluepyHelper.__init__(self)
        self._serviceMap = None  # Indexed by UUID
        (self.deviceAddr, self.addrType, self.iface) = (None, None, None)

        if isinstance(deviceAddr, ScanEntry):
            self._connect(deviceAddr.addr, deviceAddr.addrType, deviceAddr.iface, timeout)
        elif deviceAddr is not None:
            self._connect(deviceAddr, addrType, iface, timeout)

    def setDelegate(self, delegate_):  # same as withDelegate(), deprecated
        return self.withDelegate(delegate_)

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.disconnect()

    def _getResp(self, wantType, timeout=None):
        if isinstance(wantType, list) is not True:
            wantType = [wantType]

        while True:
            resp = self._waitResp(wantType + ["ntfy", "ind"], timeout)
            if resp is None:
                return None

            respType = resp["rsp"][0]
            if respType == "ntfy" or respType == "ind":
                hnd = resp["hnd"][0]
                data = resp["d"][0]
                if self.delegate is not None:
                    self.delegate.handleNotification(hnd, data)
            if respType not in wantType:
                continue
            return resp

    def _connect(self, addr, addrType=ADDR_TYPE_PUBLIC, iface=None, timeout=None):
        if len(addr.split(":")) != 6:
            raise ValueError(f"Expected MAC address, got {repr(addr)}")
        if addrType not in (ADDR_TYPE_PUBLIC, ADDR_TYPE_RANDOM):
            raise ValueError(f"Expected address type public or random, got {addrType}")
        self.retries = 3
        while self.retries > 0:
            self._startHelper(iface)
            self.addr = addr
            self.addrType = addrType
            self.iface = iface
            if iface is not None:
                self._writeCmd(f"conn {addr} {addrType} hci{str(iface)}\n")
            else:
                self._writeCmd(f"conn {addr} {addrType}\n")
            rsp = self._getResp("stat", timeout)
            timeout_exception = BTLEDisconnectError(
                f"Timed out while trying to connect to peripheral {addr}, addr type: {addrType}",
                rsp,
            )
            if rsp is None:
                raise timeout_exception
            while rsp and rsp["state"][0] == "tryconn":
                rsp = self._getResp("stat", timeout)
            if rsp is not None and rsp["state"][0] == "conn":
                DBG("   *** Succesfully connected.")
                # successful
                self.retries = 0
            if rsp is None or rsp["state"][0] != "conn":
                self._stopHelper()
                if rsp is None:
                    raise timeout_exception
                else:
                    DBG(f"   *** Failed to connect. ({self.retries})")
                    time.sleep(5.0)
                    if self.retries <= 1:
                        raise BTLEDisconnectError(
                            f"Failed to connect to peripheral {addr}, addr type: {addrType}",
                            rsp,
                        )
            self.retries -= 1

    def connect(self, addr, addrType=ADDR_TYPE_PUBLIC, iface=None, timeout=None):
        if isinstance(addr, ScanEntry):
            self._connect(addr.addr, addr.addrType, addr.iface, timeout)
        elif addr is not None:
            self._connect(addr, addrType, iface, timeout)

    def disconnect(self):
        if self._helper is None:
            return
        # Unregister the delegate first
        self.setDelegate(None)

        self._writeCmd("disc\n")
        self._getResp("stat")
        self._stopHelper()

    def discoverServices(self):
        self._writeCmd("svcs\n")
        rsp = self._getResp("find")
        starts = rsp["hstart"]
        ends = rsp["hend"]
        uuids = rsp["uuid"]
        nSvcs = len(uuids)
        assert len(starts) == nSvcs and len(ends) == nSvcs
        self._serviceMap = {}
        for i in range(nSvcs):
            self._serviceMap[UUID(uuids[i])] = Service(self, uuids[i], starts[i], ends[i])
        return self._serviceMap

    def getState(self):
        status = self.status()
        return status["state"][0]

    @property
    def services(self):
        if self._serviceMap is None:
            self._serviceMap = self.discoverServices()
        return list(self._serviceMap.values())

    def getServices(self):
        return self.services

    def getServiceByUUID(self, uuidVal):
        uuid = UUID(uuidVal)
        if self._serviceMap is not None and uuid in self._serviceMap:
            return self._serviceMap[uuid]
        self._writeCmd(f"svcs {uuid}\n")
        rsp = self._getResp("find")
        if "hstart" not in rsp:
            raise BTLEGattError(f"Service {uuid.getCommonName()} not found", rsp)
        svc = Service(self, uuid, rsp["hstart"][0], rsp["hend"][0])

        if self._serviceMap is None:
            self._serviceMap = {}
        self._serviceMap[uuid] = svc
        return svc

    def _getIncludedServices(self, startHnd=1, endHnd=0xFFFF):
        # TODO: No working example of this yet
        self._writeCmd(f"incl {startHnd:X} {endHnd:X}\n")
        return self._getResp("find")

    def getCharacteristics(self, startHnd=1, endHnd=0xFFFF, uuid=None, timeout=None):
        cmd = f"char {startHnd:X} {endHnd:X}"
        if uuid:
            cmd += f" {UUID(uuid)}"
        self._writeCmd(cmd + "\n")
        rsp = self._getResp("find", timeout)
        timeout_exception = BTLEDisconnectError(
            f"Timed out while trying to get characteristics from peripheral {self.addr}, addr type: {self.addrType}",
            rsp,
        )
        if rsp is None:
            raise timeout_exception
        nChars = len(rsp["hnd"])
        return [
            Characteristic(self, rsp["uuid"][i], rsp["hnd"][i], rsp["props"][i], rsp["vhnd"][i]) for i in range(nChars)
        ]

    def getDescriptors(self, startHnd=1, endHnd=0xFFFF):
        self._writeCmd("desc {startHnd}:X} {endHnd:X}\n")
        # Historical note:
        # Certain Bluetooth LE devices are not capable of sending back all
        # descriptors in one packet due to the limited size of MTU. So the
        # guest needs to check the response and make retries until all handles
        # are returned.
        # In bluez 5.25 and later, gatt_discover_desc() in attrib/gatt.c does the retry
        # so bluetooth_helper always returns a full list.
        # This was broken in earlier versions.
        resp = self._getResp("desc")
        ndesc = len(resp["hnd"])
        return [Descriptor(self, resp["uuid"][i], resp["hnd"][i]) for i in range(ndesc)]

    def readCharacteristic(self, handle):
        self._writeCmd(f"rd {handle:X}\n")
        resp = self._getResp("rd")
        return resp["d"][0]

    def _readCharacteristicByUUID(self, uuid, startHnd, endHnd):
        # Not used at present
        self._writeCmd(f"rdu {UUID(uuid)} {startHnd:X} {endHnd:X}\n")
        return self._getResp("rd")

    def writeCharacteristic(self, handle, val, withResponse=False, timeout=None):
        # Without response, a value too long for one packet will be truncated,
        # but with response, it will be sent as a queued write
        cmd = "wrr" if withResponse else "wr"
        bval = binascii.b2a_hex(val).decode("utf-8")
        self._writeCmd(f"{cmd} {handle:X} {bval}\n")
        return self._getResp("wr", timeout)

    def setSecurityLevel(self, level):
        self._writeCmd(f"secu {level}\n")
        return self._getResp("stat")

    def unpair(self):
        self._mgmtCmd("unpair")

    def pair(self):
        self._mgmtCmd("pair")

    def getMTU(self):
        return self._mtu

    def setMTU(self, mtu):
        self._writeCmd(f"mtu {mtu}\n")
        return self._getResp("stat")

    def waitForNotifications(self, timeout):
        resp = self._getResp(["ntfy", "ind"], timeout)
        return resp is not None

    def _setRemoteOOB(self, address, address_type, oob_data, iface=None):
        if self._helper is None:
            self._startHelper(iface)
        self.addr = address
        self.addrType = address_type
        self.iface = iface
        cmd = "remote_oob " + address + " " + address_type
        if oob_data["C_192"] is not None and oob_data["R_192"] is not None:
            cmd += " C_192 " + oob_data["C_192"] + " R_192 " + oob_data["R_192"]
        if oob_data["C_256"] is not None and oob_data["R_256"] is not None:
            cmd += " C_256 " + oob_data["C_256"] + " R_256 " + oob_data["R_256"]
        if iface is not None:
            cmd += " hci" + str(iface)
        self._writeCmd(cmd)

    def setRemoteOOB(self, address, address_type, oob_data, iface=None):
        if len(address.split(":")) != 6:
            raise ValueError(f"Expected MAC address, got {repr(address)}")
        if address_type not in (ADDR_TYPE_PUBLIC, ADDR_TYPE_RANDOM):
            raise ValueError(f"Expected address type public or random, got {address_type}")
        if isinstance(address, ScanEntry):
            return self._setOOB(address.addr, address.addrType, oob_data, address.iface)
        elif address is not None:
            return self._setRemoteOOB(address, address_type, oob_data, iface)

    def getLocalOOB(self, iface=None):
        if self._helper is None:
            self._startHelper(iface)
        self.iface = iface
        self._writeCmd("local_oob\n")
        if iface is not None:
            cmd += " hci" + str(iface)
        resp = self._getResp("oob")
        if resp is not None:
            data = resp.get("d", [""])[0]
            if data is None:
                raise BTLEManagementError("Failed to get local OOB data.")
            if struct.unpack_from("<B", data, 0)[0] != 8 or struct.unpack_from("<B", data, 1)[0] != 0x1B:
                raise BTLEManagementError("Malformed local OOB data (address).")
            address = data[2:8]
            address_type = data[8:9]
            if struct.unpack_from("<B", data, 9)[0] != 2 or struct.unpack_from("<B", data, 10)[0] != 0x1C:
                raise BTLEManagementError("Malformed local OOB data (role).")
            role = data[11:12]
            if struct.unpack_from("<B", data, 12)[0] != 17 or struct.unpack_from("<B", data, 13)[0] != 0x22:
                raise BTLEManagementError("Malformed local OOB data (confirm).")
            confirm = data[14:30]
            if struct.unpack_from("<B", data, 30)[0] != 17 or struct.unpack_from("<B", data, 31)[0] != 0x23:
                raise BTLEManagementError("Malformed local OOB data (random).")
            random = data[32:48]
            if struct.unpack_from("<B", data, 48)[0] != 2 or struct.unpack_from("<B", data, 49)[0] != 0x1:
                raise BTLEManagementError("Malformed local OOB data (flags).")
            flags = data[50:51]
            return {
                "Address": "".join(["%02X" % struct.unpack("<B", c)[0] for c in address]),
                "Type": "".join(["%02X" % struct.unpack("<B", c)[0] for c in address_type]),
                "Role": "".join(["%02X" % struct.unpack("<B", c)[0] for c in role]),
                "C_256": "".join(["%02X" % struct.unpack("<B", c)[0] for c in confirm]),
                "R_256": "".join(["%02X" % struct.unpack("<B", c)[0] for c in random]),
                "Flags": "".join(["%02X" % struct.unpack("<B", c)[0] for c in flags]),
            }

    def __del__(self):
        self.disconnect()


class ScanEntry:
    addrTypes = {1: ADDR_TYPE_PUBLIC, 2: ADDR_TYPE_RANDOM}

    FLAGS = 0x01
    INCOMPLETE_16B_SERVICES = 0x02
    COMPLETE_16B_SERVICES = 0x03
    INCOMPLETE_32B_SERVICES = 0x04
    COMPLETE_32B_SERVICES = 0x05
    INCOMPLETE_128B_SERVICES = 0x06
    COMPLETE_128B_SERVICES = 0x07
    SHORT_LOCAL_NAME = 0x08
    COMPLETE_LOCAL_NAME = 0x09
    TX_POWER = 0x0A
    SERVICE_SOLICITATION_16B = 0x14
    SERVICE_SOLICITATION_32B = 0x1F
    SERVICE_SOLICITATION_128B = 0x15
    SERVICE_DATA_16B = 0x16
    SERVICE_DATA_32B = 0x20
    SERVICE_DATA_128B = 0x21
    PUBLIC_TARGET_ADDRESS = 0x17
    RANDOM_TARGET_ADDRESS = 0x18
    APPEARANCE = 0x19
    ADVERTISING_INTERVAL = 0x1A
    MANUFACTURER = 0xFF

    dataTags = {
        FLAGS: "Flags",
        INCOMPLETE_16B_SERVICES: "Incomplete 16b Services",
        COMPLETE_16B_SERVICES: "Complete 16b Services",
        INCOMPLETE_32B_SERVICES: "Incomplete 32b Services",
        COMPLETE_32B_SERVICES: "Complete 32b Services",
        INCOMPLETE_128B_SERVICES: "Incomplete 128b Services",
        COMPLETE_128B_SERVICES: "Complete 128b Services",
        SHORT_LOCAL_NAME: "Short Local Name",
        COMPLETE_LOCAL_NAME: "Complete Local Name",
        TX_POWER: "Tx Power",
        SERVICE_SOLICITATION_16B: "16b Service Solicitation",
        SERVICE_SOLICITATION_32B: "32b Service Solicitation",
        SERVICE_SOLICITATION_128B: "128b Service Solicitation",
        SERVICE_DATA_16B: "16b Service Data",
        SERVICE_DATA_32B: "32b Service Data",
        SERVICE_DATA_128B: "128b Service Data",
        PUBLIC_TARGET_ADDRESS: "Public Target Address",
        RANDOM_TARGET_ADDRESS: "Random Target Address",
        APPEARANCE: "Appearance",
        ADVERTISING_INTERVAL: "Advertising Interval",
        MANUFACTURER: "Manufacturer",
    }

    def __init__(self, addr, iface):
        self.addr = addr
        self.iface = iface
        self.addrType = None
        self.rssi = None
        self.connectable = False
        self.rawData = None
        self.scanData = {}
        self.updateCount = 0

    def _update(self, resp):
        addrType = self.addrTypes.get(resp["type"][0], None)
        if (self.addrType is not None) and (addrType != self.addrType):
            raise BTLEInternalError(f"Address type changed during scan, for address {self.addr}")
        self.addrType = addrType
        self.rssi = -resp["rssi"][0]
        self.connectable = (resp["flag"][0] & 0x4) == 0
        data = resp.get("d", [""])[0]
        self.rawData = data

        # Note: bluez is notifying devices twice: once with advertisement data,
        # then with scan response data. Also, the device may update the
        # advertisement or scan data
        isNewData = False
        while len(data) >= 2:
            sdlen, sdid = struct.unpack_from("<BB", data)
            val = data[2 : sdlen + 1]
            if (sdid not in self.scanData) or (val != self.scanData[sdid]):
                isNewData = True
            self.scanData[sdid] = val
            data = data[sdlen + 1 :]

        self.updateCount += 1
        return isNewData

    def _decodeUUID(self, val, nbytes):
        if len(val) < nbytes:
            return None
        bval = bytearray(val)
        rs = ""
        # Bytes are little-endian; convert to big-endian string
        for i in range(nbytes):
            rs = f"{bval[i]:02X}{rs}"
        return UUID(rs)

    def _decodeUUIDlist(self, val, nbytes):
        result = []
        for i in range(0, len(val), nbytes):
            if len(val) >= (i + nbytes):
                result.append(self._decodeUUID(val[i : i + nbytes], nbytes))
        return result

    def getDescription(self, sdid):
        return self.dataTags.get(sdid, hex(sdid))

    def getValue(self, sdid):
        val = self.scanData.get(sdid, None)
        if val is None:
            return None
        if sdid in [ScanEntry.SHORT_LOCAL_NAME, ScanEntry.COMPLETE_LOCAL_NAME]:
            try:
                # Beware! Vol 3 Part C 18.3 doesn't give an encoding. Other references
                # to 'local name' (e.g. vol 3 E, 6.23) suggest it's UTF-8 but in practice
                # devices sometimes have garbage here. See #259, #275, #292.
                return val.decode("utf-8")
            except UnicodeDecodeError:
                bbval = bytearray(val)
                return "".join([(chr(x) if (x >= 32 and x <= 127) else "?") for x in bbval])
        elif sdid in [
            ScanEntry.INCOMPLETE_16B_SERVICES,
            ScanEntry.COMPLETE_16B_SERVICES,
        ]:
            return self._decodeUUIDlist(val, 2)
        elif sdid in [
            ScanEntry.INCOMPLETE_32B_SERVICES,
            ScanEntry.COMPLETE_32B_SERVICES,
        ]:
            return self._decodeUUIDlist(val, 4)
        elif sdid in [
            ScanEntry.INCOMPLETE_128B_SERVICES,
            ScanEntry.COMPLETE_128B_SERVICES,
        ]:
            return self._decodeUUIDlist(val, 16)
        else:
            return val

    def getValueText(self, sdid):
        val = self.getValue(sdid)
        if val is None:
            return None
        if sdid in [ScanEntry.SHORT_LOCAL_NAME, ScanEntry.COMPLETE_LOCAL_NAME]:
            return val
        elif isinstance(val, list):
            return ",".join(str(v) for v in val)
        else:
            return binascii.b2a_hex(val).decode("ascii")

    def getScanData(self):
        """Return list of tuples [(tag, description, value)]"""
        return [(sdid, self.getDescription(sdid), self.getValueText(sdid)) for sdid in self.scanData.keys()]


class Scanner(BluepyHelper):
    def __init__(self, iface=0):
        BluepyHelper.__init__(self)
        self.scanned = {}
        self.iface = iface
        self.passive = False

    def _cmd(self):
        return "pasv" if self.passive else "scan"

    def start(self, passive=False):
        self.passive = passive
        self._startHelper(iface=self.iface)
        self._mgmtCmd("le on")
        self._writeCmd(self._cmd() + "\n")
        rsp = self._waitResp("mgmt")
        if rsp["code"][0] == "success":
            return
        # Sometimes previous scan still ongoing
        if rsp["code"][0] == "busy":
            self._mgmtCmd(self._cmd() + "end")
            rsp = self._waitResp("stat")
            assert rsp["state"][0] == "disc"
            self._mgmtCmd(self._cmd())

    def stop(self):
        self._mgmtCmd(self._cmd() + "end")
        self._stopHelper()

    def clear(self):
        self.scanned = {}

    def process(self, timeout=10.0):
        if self._helper is None:
            raise BTLEInternalError("Helper not started (did you call start()?)")
        start = time.time()
        while True:
            if timeout:
                remain = start + timeout - time.time()
                if remain <= 0.0:
                    break
            else:
                remain = None
            resp = self._waitResp(["scan", "stat"], remain)
            if resp is None:
                break

            respType = resp["rsp"][0]
            if respType == "stat":
                # if scan ended, restart it
                if resp["state"][0] == "disc":
                    self._mgmtCmd(self._cmd())

            elif respType == "scan":
                # device found
                addr = binascii.b2a_hex(resp["addr"][0]).decode("utf-8")
                addr = ":".join([addr[i : i + 2] for i in range(0, 12, 2)])
                if addr in self.scanned:
                    dev = self.scanned[addr]
                else:
                    dev = ScanEntry(addr, self.iface)
                    self.scanned[addr] = dev
                isNewData = dev._update(resp)
                if self.delegate is not None:
                    self.delegate.handleDiscovery(dev, (dev.updateCount <= 1), isNewData)

            else:
                raise BTLEInternalError(f"Unexpected response: {respType}", resp)

    def getDevices(self):
        return list(self.scanned.values())

    def scan(self, timeout=10, passive=False):
        self.clear()
        self.start(passive=passive)
        self.process(timeout)
        self.stop()
        return self.getDevices()


def capitaliseName(descr):
    words = descr.replace("(", " ").replace(")", " ").replace("-", " ").split(" ")
    capWords = [words[0].lower()]
    capWords += [w[0:1].upper() + w[1:].lower() for w in words[1:]]
    return "".join(capWords)


class _UUIDNameMap:
    # Constructor sets self.currentTimeService, self.txPower, and so on
    # from names.
    def __init__(self, idList):
        self.idMap = {}

        for uuid in idList:
            attrName = capitaliseName(uuid.commonName)
            vars(self)[attrName] = uuid
            self.idMap[uuid] = uuid

    def getCommonName(self, uuid):
        if uuid in self.idMap:
            return self.idMap[uuid].commonName
        return None


def get_json_uuid():
    import json

    with open(os.path.join(script_path, "uuids.json"), "rb") as fp:
        uuid_data = json.loads(fp.read().decode("utf-8"))
    for k in uuid_data.keys():
        for number, cname, name in uuid_data[k]:
            yield UUID(number, cname)
            yield UUID(number, name)


AssignedNumbers = _UUIDNameMap(get_json_uuid())

if __name__ == "__main__":
    if len(sys.argv) < 2:
        sys.exit(f"Usage:\n  {sys.argv[0]} <mac-address> [random]")

    if not os.path.isfile(helperExe):
        raise ImportError(f"Cannot find required executable '{helperExe}'")

    devAddr = sys.argv[1]
    if len(sys.argv) == 3:
        addrType = sys.argv[2]
    else:
        addrType = ADDR_TYPE_PUBLIC
    print(f"Connecting to: {devAddr}, address type: {addrType}")
    conn = Peripheral(devAddr, addrType)
    try:
        for svc in conn.services:
            print(f"{str(svc)}:")
            for ch in svc.getCharacteristics():
                print(f"    {ch}, hnd={hex(ch.handle)}, supports {ch.propertiesToString()}")
                chName = AssignedNumbers.getCommonName(ch.uuid)
                if ch.supportsRead():
                    try:
                        print(f"    ->{repr(ch.read())}")
                    except BTLEException as e:
                        print(f"    ->{e}")

    finally:
        conn.disconnect()



#!/usr/bin/env python3

import argparse
import binascii
import time

from .btle import UUID, Peripheral, ADDR_TYPE_RANDOM, DefaultDelegate


def write_uint16(data, value, index):
    """Write 16bit value into data string at index and return new string"""
    data = data.decode("utf-8")  # This line is added to make sure both Python 2 and 3 works
    return f"{data[: index * 4]}{value & 0xFF:02x}{value >> 8:02x}{data[index * 4 + 4 :]}"


def write_uint8(data, value, index):
    """Write 8bit value into data string at index and return new string"""
    data = data.decode("utf-8")  # This line is added to make sure both Python 2 and 3 works
    return f"{data[: index * 2]}{value:02x}{data[index * 2 + 2 :]}"


# Please see # Ref https://nordicsemiconductor.github.io/Nordic-Thingy52-FW/documentation
# for more information on the UUIDs of the Services and Characteristics that are being used
def Nordic_UUID(val):
    """Adds base UUID and inserts value to return Nordic UUID"""
    return UUID(f"EF68{val:04X}-9B35-4933-9B10-52FFA9740042")


# Definition of all UUID used by Thingy
CCCD_UUID = 0x2902

BATTERY_SERVICE_UUID = 0x180F
BATTERY_LEVEL_UUID = 0x2A19

ENVIRONMENT_SERVICE_UUID = 0x0200
E_TEMPERATURE_CHAR_UUID = 0x0201
E_PRESSURE_CHAR_UUID = 0x0202
E_HUMIDITY_CHAR_UUID = 0x0203
E_GAS_CHAR_UUID = 0x0204
E_COLOR_CHAR_UUID = 0x0205
E_CONFIG_CHAR_UUID = 0x0206

USER_INTERFACE_SERVICE_UUID = 0x0300
UI_LED_CHAR_UUID = 0x0301
UI_BUTTON_CHAR_UUID = 0x0302
UI_EXT_PIN_CHAR_UUID = 0x0303

MOTION_SERVICE_UUID = 0x0400
M_CONFIG_CHAR_UUID = 0x0401
M_TAP_CHAR_UUID = 0x0402
M_ORIENTATION_CHAR_UUID = 0x0403
M_QUATERNION_CHAR_UUID = 0x0404
M_STEP_COUNTER_UUID = 0x0405
M_RAW_DATA_CHAR_UUID = 0x0406
M_EULER_CHAR_UUID = 0x0407
M_ROTATION_MATRIX_CHAR_UUID = 0x0408
M_HEAIDNG_CHAR_UUID = 0x0409
M_GRAVITY_VECTOR_CHAR_UUID = 0x040A

SOUND_SERVICE_UUID = 0x0500
S_CONFIG_CHAR_UUID = 0x0501
S_SPEAKER_DATA_CHAR_UUID = 0x0502
S_SPEAKER_STATUS_CHAR_UUID = 0x0503
S_MICROPHONE_CHAR_UUID = 0x0504

# Notification handles used in notification delegate
e_temperature_handle = None
e_pressure_handle = None
e_humidity_handle = None
e_gas_handle = None
e_color_handle = None
ui_button_handle = None
m_tap_handle = None
m_orient_handle = None
m_quaternion_handle = None
m_stepcnt_handle = None
m_rawdata_handle = None
m_euler_handle = None
m_rotation_handle = None
m_heading_handle = None
m_gravity_handle = None
s_speaker_status_handle = None
s_microphone_handle = None


class BatterySensor:
    """
    Battery Service module. Instance the class and enable to get access to Battery interface.
    """

    svcUUID = UUID(BATTERY_SERVICE_UUID)  # Ref https://www.bluetooth.com/specifications/gatt/services
    dataUUID = UUID(BATTERY_LEVEL_UUID)  # Ref https://www.bluetooth.com/specifications/gatt/characteristics

    def __init__(self, periph):
        self.periph = periph
        self.service = None
        self.data = None

    def enable(self):
        """Enables the class by finding the service and its characteristics."""
        if self.service is None:
            self.service = self.periph.getServiceByUUID(self.svcUUID)
        if self.data is None:
            self.data = self.service.getCharacteristics(self.dataUUID)[0]

    def read(self):
        """Returns the battery level in percent"""
        val = ord(self.data.read())
        return val


class EnvironmentService:
    """
    Environment service module. Instance the class and enable to get access to the Environment interface.
    """

    serviceUUID = Nordic_UUID(ENVIRONMENT_SERVICE_UUID)
    temperature_char_uuid = Nordic_UUID(E_TEMPERATURE_CHAR_UUID)
    pressure_char_uuid = Nordic_UUID(E_PRESSURE_CHAR_UUID)
    humidity_char_uuid = Nordic_UUID(E_HUMIDITY_CHAR_UUID)
    gas_char_uuid = Nordic_UUID(E_GAS_CHAR_UUID)
    color_char_uuid = Nordic_UUID(E_COLOR_CHAR_UUID)
    config_char_uuid = Nordic_UUID(E_CONFIG_CHAR_UUID)

    def __init__(self, periph):
        self.periph = periph
        self.environment_service = None
        self.temperature_char = None
        self.temperature_cccd = None
        self.pressure_char = None
        self.pressure_cccd = None
        self.humidity_char = None
        self.humidity_cccd = None
        self.gas_char = None
        self.gas_cccd = None
        self.color_char = None
        self.color_cccd = None
        self.config_char = None

    def enable(self):
        """Enables the class by finding the service and its characteristics."""
        global e_temperature_handle
        global e_pressure_handle
        global e_humidity_handle
        global e_gas_handle
        global e_color_handle

        if self.environment_service is None:
            self.environment_service = self.periph.getServiceByUUID(self.serviceUUID)
        if self.temperature_char is None:
            self.temperature_char = self.environment_service.getCharacteristics(self.temperature_char_uuid)[0]
            e_temperature_handle = self.temperature_char.getHandle()
            self.temperature_cccd = self.temperature_char.getDescriptors(forUUID=CCCD_UUID)[0]
        if self.pressure_char is None:
            self.pressure_char = self.environment_service.getCharacteristics(self.pressure_char_uuid)[0]
            e_pressure_handle = self.pressure_char.getHandle()
            self.pressure_cccd = self.pressure_char.getDescriptors(forUUID=CCCD_UUID)[0]
        if self.humidity_char is None:
            self.humidity_char = self.environment_service.getCharacteristics(self.humidity_char_uuid)[0]
            e_humidity_handle = self.humidity_char.getHandle()
            self.humidity_cccd = self.humidity_char.getDescriptors(forUUID=CCCD_UUID)[0]
        if self.gas_char is None:
            self.gas_char = self.environment_service.getCharacteristics(self.gas_char_uuid)[0]
            e_gas_handle = self.gas_char.getHandle()
            self.gas_cccd = self.gas_char.getDescriptors(forUUID=CCCD_UUID)[0]
        if self.color_char is None:
            self.color_char = self.environment_service.getCharacteristics(self.color_char_uuid)[0]
            e_color_handle = self.color_char.getHandle()
            self.color_cccd = self.color_char.getDescriptors(forUUID=CCCD_UUID)[0]
        if self.config_char is None:
            self.config_char = self.environment_service.getCharacteristics(self.config_char_uuid)[0]

    def set_temperature_notification(self, state):
        if self.temperature_cccd is not None:
            if state is True:
                self.temperature_cccd.write(b"\x01\x00", True)
            else:
                self.temperature_cccd.write(b"\x00\x00", True)

    def set_pressure_notification(self, state):
        if self.pressure_cccd is not None:
            if state is True:
                self.pressure_cccd.write(b"\x01\x00", True)
            else:
                self.pressure_cccd.write(b"\x00\x00", True)

    def set_humidity_notification(self, state):
        if self.humidity_cccd is not None:
            if state is True:
                self.humidity_cccd.write(b"\x01\x00", True)
            else:
                self.humidity_cccd.write(b"\x00\x00", True)

    def set_gas_notification(self, state):
        if self.gas_cccd is not None:
            if state is True:
                self.gas_cccd.write(b"\x01\x00", True)
            else:
                self.gas_cccd.write(b"\x00\x00", True)

    def set_color_notification(self, state):
        if self.color_cccd is not None:
            if state is True:
                self.color_cccd.write(b"\x01\x00", True)
            else:
                self.color_cccd.write(b"\x00\x00", True)

    # fmt: off
    def configure(self, temp_int=None, press_int=None, humid_int=None, gas_mode_int=None, color_int=None, color_sens_calib=None,):
        # fmt: on
        if self.config_char is not None:
            current_config = binascii.b2a_hex(self.config_char.read())
            if temp_int is not None:
                new_config = write_uint16(current_config, temp_int, 0)
            if press_int is not None:
                new_config = write_uint16(current_config, press_int, 1)
            if humid_int is not None:
                new_config = write_uint16(current_config, humid_int, 2)
            if gas_mode_int is not None:
                new_config = write_uint8(current_config, gas_mode_int, 8)
            if color_int is not None:
                new_config = write_uint16(current_config, color_int, 3)
            if color_sens_calib is not None:
                new_config = write_uint8(current_config, color_sens_calib[0], 9)
                new_config = write_uint8(current_config, color_sens_calib[1], 10)
                new_config = write_uint8(current_config, color_sens_calib[2], 11)
            self.config_char.write(binascii.a2b_hex(new_config), True)

    def disable(self):
        self.set_temperature_notification(False)
        self.set_pressure_notification(False)
        self.set_humidity_notification(False)
        self.set_gas_notification(False)
        self.set_color_notification(False)


class UserInterfaceService:
    """
    User interface service module. Instance the class and enable to get access to the UI interface.
    """

    serviceUUID = Nordic_UUID(USER_INTERFACE_SERVICE_UUID)
    led_char_uuid = Nordic_UUID(UI_LED_CHAR_UUID)
    btn_char_uuid = Nordic_UUID(UI_BUTTON_CHAR_UUID)

    # To be added: EXT PIN CHAR

    def __init__(self, periph):
        self.periph = periph
        self.ui_service = None
        self.led_char = None
        self.btn_char = None
        self.btn_char_cccd = None  # To be added: EXT PIN CHAR

    def enable(self):
        """Enables the class by finding the service and its characteristics."""
        global ui_button_handle

        if self.ui_service is None:
            self.ui_service = self.periph.getServiceByUUID(self.serviceUUID)
        if self.led_char is None:
            self.led_char = self.ui_service.getCharacteristics(self.led_char_uuid)[0]
        if self.btn_char is None:
            self.btn_char = self.ui_service.getCharacteristics(self.btn_char_uuid)[0]
            ui_button_handle = self.btn_char.getHandle()
            self.btn_char_cccd = self.btn_char.getDescriptors(forUUID=CCCD_UUID)[0]

    def set_led_mode_off(self):
        self.led_char.write(b"\x00", True)

    def set_led_mode_constant(self, r, g, b):
        teptep = f"01{r:02X}{g:02X}{b:02X}"
        self.led_char.write(binascii.a2b_hex(teptep), True)

    def set_led_mode_breathe(self, color, intensity, delay):
        """
        Set LED to breathe mode.
        color has to be within 0x01 and 0x07
        intensity [%] has to be within 1-100
        delay [ms] has to be within 1 ms - 10 s
        """
        teptep = f"02{color:02X}{intensity:02X}{delay & 0xFF:02X}{delay >> 8:02X}"
        self.led_char.write(binascii.a2b_hex(teptep), True)

    def set_led_mode_one_shot(self, color, intensity):
        """
        Set LED to one shot mode.
        color has to be within 0x01 and 0x07
        intensity [%] has to be within 1-100
        """
        teptep = "03{color:02X}{intensity:02X}"
        self.led_char.write(binascii.a2b_hex(teptep), True)

    def set_btn_notification(self, state):
        if self.btn_char_cccd is not None:
            if state is True:
                self.btn_char_cccd.write(b"\x01\x00", True)
            else:
                self.btn_char_cccd.write(b"\x00\x00", True)

    def disable(self):
        self.set_btn_notification(False)


class MotionService:
    """
    Motion service module. Instance the class and enable to get access to the Motion interface.
    """

    serviceUUID = Nordic_UUID(MOTION_SERVICE_UUID)
    config_char_uuid = Nordic_UUID(M_CONFIG_CHAR_UUID)
    tap_char_uuid = Nordic_UUID(M_TAP_CHAR_UUID)
    orient_char_uuid = Nordic_UUID(M_ORIENTATION_CHAR_UUID)
    quaternion_char_uuid = Nordic_UUID(M_QUATERNION_CHAR_UUID)
    stepcnt_char_uuid = Nordic_UUID(M_STEP_COUNTER_UUID)
    rawdata_char_uuid = Nordic_UUID(M_RAW_DATA_CHAR_UUID)
    euler_char_uuid = Nordic_UUID(M_EULER_CHAR_UUID)
    rotation_char_uuid = Nordic_UUID(M_ROTATION_MATRIX_CHAR_UUID)
    heading_char_uuid = Nordic_UUID(M_HEAIDNG_CHAR_UUID)
    gravity_char_uuid = Nordic_UUID(M_GRAVITY_VECTOR_CHAR_UUID)

    def __init__(self, periph):
        self.periph = periph
        self.motion_service = None
        self.config_char = None
        self.tap_char = None
        self.tap_char_cccd = None
        self.orient_char = None
        self.orient_cccd = None
        self.quaternion_char = None
        self.quaternion_cccd = None
        self.stepcnt_char = None
        self.stepcnt_cccd = None
        self.rawdata_char = None
        self.rawdata_cccd = None
        self.euler_char = None
        self.euler_cccd = None
        self.rotation_char = None
        self.rotation_cccd = None
        self.heading_char = None
        self.heading_cccd = None
        self.gravity_char = None
        self.gravity_cccd = None

    def enable(self):
        """Enables the class by finding the service and its characteristics."""
        global m_tap_handle
        global m_orient_handle
        global m_quaternion_handle
        global m_stepcnt_handle
        global m_rawdata_handle
        global m_euler_handle
        global m_rotation_handle
        global m_heading_handle
        global m_gravity_handle

        if self.motion_service is None:
            self.motion_service = self.periph.getServiceByUUID(self.serviceUUID)
        if self.config_char is None:
            self.config_char = self.motion_service.getCharacteristics(self.config_char_uuid)[0]
        if self.tap_char is None:
            self.tap_char = self.motion_service.getCharacteristics(self.tap_char_uuid)[0]
            m_tap_handle = self.tap_char.getHandle()
            self.tap_char_cccd = self.tap_char.getDescriptors(forUUID=CCCD_UUID)[0]
        if self.orient_char is None:
            self.orient_char = self.motion_service.getCharacteristics(self.orient_char_uuid)[0]
            m_orient_handle = self.orient_char.getHandle()
            self.orient_cccd = self.orient_char.getDescriptors(forUUID=CCCD_UUID)[0]
        if self.quaternion_char is None:
            self.quaternion_char = self.motion_service.getCharacteristics(self.quaternion_char_uuid)[0]
            m_quaternion_handle = self.quaternion_char.getHandle()
            self.quaternion_cccd = self.quaternion_char.getDescriptors(forUUID=CCCD_UUID)[0]
        if self.stepcnt_char is None:
            self.stepcnt_char = self.motion_service.getCharacteristics(self.stepcnt_char_uuid)[0]
            m_stepcnt_handle = self.stepcnt_char.getHandle()
            self.stepcnt_cccd = self.stepcnt_char.getDescriptors(forUUID=CCCD_UUID)[0]
        if self.rawdata_char is None:
            self.rawdata_char = self.motion_service.getCharacteristics(self.rawdata_char_uuid)[0]
            m_rawdata_handle = self.rawdata_char.getHandle()
            self.rawdata_cccd = self.rawdata_char.getDescriptors(forUUID=CCCD_UUID)[0]
        if self.euler_char is None:
            self.euler_char = self.motion_service.getCharacteristics(self.euler_char_uuid)[0]
            m_euler_handle = self.euler_char.getHandle()
            self.euler_cccd = self.euler_char.getDescriptors(forUUID=CCCD_UUID)[0]
        if self.rotation_char is None:
            self.rotation_char = self.motion_service.getCharacteristics(self.rotation_char_uuid)[0]
            m_rotation_handle = self.rotation_char.getHandle()
            self.rotation_cccd = self.rotation_char.getDescriptors(forUUID=CCCD_UUID)[0]
        if self.heading_char is None:
            self.heading_char = self.motion_service.getCharacteristics(self.heading_char_uuid)[0]
            m_heading_handle = self.heading_char.getHandle()
            self.heading_cccd = self.heading_char.getDescriptors(forUUID=CCCD_UUID)[0]
        if self.gravity_char is None:
            self.gravity_char = self.motion_service.getCharacteristics(self.gravity_char_uuid)[0]
            m_gravity_handle = self.gravity_char.getHandle()
            self.gravity_cccd = self.gravity_char.getDescriptors(forUUID=CCCD_UUID)[0]

    def set_tap_notification(self, state):
        if self.tap_char_cccd is not None:
            if state is True:
                self.tap_char_cccd.write(b"\x01\x00", True)
            else:
                self.tap_char_cccd.write(b"\x00\x00", True)

    def set_orient_notification(self, state):
        if self.orient_cccd is not None:
            if state is True:
                self.orient_cccd.write(b"\x01\x00", True)
            else:
                self.orient_cccd.write(b"\x00\x00", True)

    def set_quaternion_notification(self, state):
        if self.quaternion_cccd is not None:
            if state is True:
                self.quaternion_cccd.write(b"\x01\x00", True)
            else:
                self.quaternion_cccd.write(b"\x00\x00", True)

    def set_stepcnt_notification(self, state):
        if self.stepcnt_cccd is not None:
            if state is True:
                self.stepcnt_cccd.write(b"\x01\x00", True)
            else:
                self.stepcnt_cccd.write(b"\x00\x00", True)

    def set_rawdata_notification(self, state):
        if self.rawdata_cccd is not None:
            if state is True:
                self.rawdata_cccd.write(b"\x01\x00", True)
            else:
                self.rawdata_cccd.write(b"\x00\x00", True)

    def set_euler_notification(self, state):
        if self.euler_cccd is not None:
            if state is True:
                self.euler_cccd.write(b"\x01\x00", True)
            else:
                self.euler_cccd.write(b"\x00\x00", True)

    def set_rotation_notification(self, state):
        if self.rotation_cccd is not None:
            if state is True:
                self.rotation_cccd.write(b"\x01\x00", True)
            else:
                self.rotation_cccd.write(b"\x00\x00", True)

    def set_heading_notification(self, state):
        if self.heading_cccd is not None:
            if state is True:
                self.heading_cccd.write(b"\x01\x00", True)
            else:
                self.heading_cccd.write(b"\x00\x00", True)

    def set_gravity_notification(self, state):
        if self.gravity_cccd is not None:
            if state is True:
                self.gravity_cccd.write(b"\x01\x00", True)
            else:
                self.gravity_cccd.write(b"\x00\x00", True)

    def configure(
        self,
        step_int=None,
        temp_comp_int=None,
        magnet_comp_int=None,
        motion_freq=None,
        wake_on_motion=None,
    ):
        if step_int is not None and self.config_char is not None:
            current_config = binascii.b2a_hex(self.config_char.read())
            new_config = write_uint16(current_config, step_int, 0)
            self.config_char.write(binascii.a2b_hex(new_config), True)
        if temp_comp_int is not None and self.config_char is not None:
            current_config = binascii.b2a_hex(self.config_char.read())
            new_config = write_uint16(current_config, temp_comp_int, 1)
            self.config_char.write(binascii.a2b_hex(new_config), True)
        if magnet_comp_int is not None and self.config_char is not None:
            current_config = binascii.b2a_hex(self.config_char.read())
            new_config = write_uint16(current_config, magnet_comp_int, 2)
            self.config_char.write(binascii.a2b_hex(new_config), True)
        if motion_freq is not None and self.config_char is not None:
            current_config = binascii.b2a_hex(self.config_char.read())
            new_config = write_uint16(current_config, motion_freq, 3)
            self.config_char.write(binascii.a2b_hex(new_config), True)
        if wake_on_motion is not None and self.config_char is not None:
            current_config = binascii.b2a_hex(self.config_char.read())
            new_config = write_uint8(current_config, wake_on_motion, 8)
            self.config_char.write(binascii.a2b_hex(new_config), True)

    def disable(self):
        self.set_tap_notification(False)
        self.set_orient_notification(False)
        self.set_quaternion_notification(False)
        self.set_stepcnt_notification(False)
        self.set_rawdata_notification(False)
        self.set_euler_notification(False)
        self.set_rotation_notification(False)
        self.set_heading_notification(False)
        self.set_gravity_notification(False)


class SoundService:
    """
    Sound service module. Instance the class and enable to get access to the Sound interface.
    """

    serviceUUID = Nordic_UUID(SOUND_SERVICE_UUID)
    config_char_uuid = Nordic_UUID(S_CONFIG_CHAR_UUID)
    speaker_data_char_uuid = Nordic_UUID(S_SPEAKER_DATA_CHAR_UUID)
    speaker_status_char_uuid = Nordic_UUID(S_SPEAKER_STATUS_CHAR_UUID)
    microphone_char_uuid = Nordic_UUID(S_MICROPHONE_CHAR_UUID)

    def __init__(self, periph):
        self.periph = periph
        self.sound_service = None
        self.config_char = None
        self.speaker_data_char = None
        self.speaker_status_char = None
        self.speaker_status_char_cccd = None
        self.microphone_char = None
        self.microphone_char_cccd = None

    def enable(self):
        """Enables the class by finding the service and its characteristics."""
        global s_speaker_status_handle
        global s_microphone_handle

        if self.sound_service is None:
            self.sound_service = self.periph.getServiceByUUID(self.serviceUUID)
        if self.config_char is None:
            self.config_char = self.sound_service.getCharacteristics(self.config_char_uuid)[0]
        if self.speaker_data_char is None:
            self.speaker_data_char = self.sound_service.getCharacteristics(self.speaker_data_char_uuid)[0]
        if self.speaker_status_char is None:
            self.speaker_status_char = self.sound_service.getCharacteristics(self.speaker_status_char_uuid)[0]
            s_speaker_status_handle = self.speaker_status_char.getHandle()
            self.speaker_status_char_cccd = self.speaker_status_char.getDescriptors(forUUID=CCCD_UUID)[0]
        if self.microphone_char is None:
            self.microphone_char = self.sound_service.getCharacteristics(self.microphone_char_uuid)[0]
            s_microphone_handle = self.microphone_char.getHandle()
            self.microphone_char_cccd = self.microphone_char.getDescriptors(forUUID=CCCD_UUID)[0]

    def play_speaker_sample(self, sample=0):
        if self.speaker_data_char is not None:
            sample_str = "f{sample:02X}"
            self.speaker_data_char.write(binascii.a2b_hex(sample_str), False)

    def set_speaker_status_notification(self, state):
        if self.speaker_status_char_cccd is not None:
            if state is True:
                self.speaker_status_char_cccd.write(b"\x01\x00", True)
            else:
                self.speaker_status_char_cccd.write(b"\x00\x00", True)

    def set_microphone_notification(self, state):
        if self.microphone_char_cccd is not None:
            if state is True:
                self.microphone_char_cccd.write(b"\x01\x00", True)
            else:
                self.microphone_char_cccd.write(b"\x00\x00", True)

    def configure(self, speaker_mode=None, microphone_mode=None):
        if speaker_mode is not None and self.config_char is not None:
            current_config = binascii.b2a_hex(self.config_char.read())
            new_config = write_uint8(current_config, speaker_mode, 0)
            self.config_char.write(binascii.a2b_hex(new_config), True)
        if microphone_mode is not None and self.config_char is not None:
            current_config = binascii.b2a_hex(self.config_char.read())
            new_config = write_uint8(current_config, microphone_mode, 1)
            self.config_char.write(binascii.a2b_hex(new_config), True)

    def disable(self):
        self.set_speaker_status_notification(False)
        self.set_microphone_notification(False)


class MyDelegate(DefaultDelegate):
    def handleNotification(self, hnd, data):
        # Debug print repr(data)
        if hnd == e_temperature_handle:
            teptep = binascii.b2a_hex(data)
            print(f"Notification: Temp received:  {self._str_to_int(teptep[:-2])}.{int(teptep[-2:], 16)} degCelsius")

        elif hnd == e_pressure_handle:
            pressure_int, pressure_dec = self._extract_pressure_data(data)
            print(f"Notification: Press received: {pressure_int}.{pressure_dec} hPa")

        elif hnd == e_humidity_handle:
            teptep = binascii.b2a_hex(data)
            print(f"Notification: Humidity received: {self._str_to_int(teptep)} %")

        elif hnd == e_gas_handle:
            eco2, tvoc = self._extract_gas_data(data)
            print(f"Notification: Gas received: eCO2 ppm: {eco2}, TVOC ppb: {tvoc} %")

        elif hnd == e_color_handle:
            teptep = binascii.b2a_hex(data)
            print(f"Notification: Color: {teptep}")

        elif hnd == ui_button_handle:
            teptep = binascii.b2a_hex(data)
            print(f"Notification: Button state [1 -> released]: {self._str_to_int(teptep)}")

        elif hnd == m_tap_handle:
            direction, count = self._extract_tap_data(data)
            print(f"Notification: Tap: direction: {direction}, count: {self._str_to_int(count)}")

        elif hnd == m_orient_handle:
            teptep = binascii.b2a_hex(data)
            print(f"Notification: Orient: {teptep}")

        elif hnd == m_quaternion_handle:
            teptep = binascii.b2a_hex(data)
            print(f"Notification: Quaternion: {teptep}")

        elif hnd == m_stepcnt_handle:
            teptep = binascii.b2a_hex(data)
            print(f"Notification: Step Count: {teptep}")

        elif hnd == m_rawdata_handle:
            teptep = binascii.b2a_hex(data)
            print(f"Notification: Raw data: {teptep}")

        elif hnd == m_euler_handle:
            teptep = binascii.b2a_hex(data)
            print(f"Notification: Euler: {teptep}")

        elif hnd == m_rotation_handle:
            teptep = binascii.b2a_hex(data)
            print(f"Notification: Rotation matrix: {teptep}")

        elif hnd == m_heading_handle:
            teptep = binascii.b2a_hex(data)
            print(f"Notification: Heading: {teptep}")

        elif hnd == m_gravity_handle:
            teptep = binascii.b2a_hex(data)
            print(f"Notification: Gravity: {teptep}")

        elif hnd == s_speaker_status_handle:
            teptep = binascii.b2a_hex(data)
            print(f"Notification: Speaker Status: {teptep}")

        elif hnd == s_microphone_handle:
            teptep = binascii.b2a_hex(data)
            print(f"Notification: Microphone: {teptep}")

        else:
            teptep = binascii.b2a_hex(data)
            print(f"Notification: UNKOWN: hnd {hnd}, data {teptep}")

    def _str_to_int(self, s):
        """Transform hex str into int."""
        i = int(s, 16)
        if i >= 2**7:
            i -= 2**8
        return i

    def _extract_pressure_data(self, data):
        """Extract pressure data from data string."""
        teptep = binascii.b2a_hex(data)
        pressure_int = 0
        for i in range(0, 4):
            pressure_int += int(teptep[i * 2 : (i * 2) + 2], 16) << 8 * i
        pressure_dec = int(teptep[-2:], 16)
        return (pressure_int, pressure_dec)

    def _extract_gas_data(self, data):
        """Extract gas data from data string."""
        teptep = binascii.b2a_hex(data)
        eco2 = int(teptep[:2], 16) + (int(teptep[2:4], 16) << 8)
        tvoc = int(teptep[4:6], 16) + (int(teptep[6:8], 16) << 8)
        return eco2, tvoc

    def _extract_tap_data(self, data):
        """Extract tap data from data string."""
        teptep = binascii.b2a_hex(data)
        direction = teptep[0:2]
        count = teptep[2:4]
        return (direction, count)


class Thingy52(Peripheral):
    """
    Thingy:52 module. Instance the class and enable to get access to the Thingy:52 Sensors.
    The addr of your device has to be know, or can be found by using the hcitool command line
    tool, for example. Call "> sudo hcitool lescan" and your Thingy's address should show up.
    """

    def __init__(self, addr):
        Peripheral.__init__(self, addr, addrType=ADDR_TYPE_RANDOM)

        # Thingy configuration service not implemented
        self.battery = BatterySensor(self)
        self.environment = EnvironmentService(self)
        self.ui = UserInterfaceService(self)
        self.motion = MotionService(self)
        self.sound = SoundService(self)  # DFU Service not implemented


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("mac_address", action="store", help="MAC address of BLE peripheral")
    parser.add_argument(
        "-n",
        action="store",
        dest="count",
        default=0,
        type=int,
        help="Number of times to loop data",
    )
    parser.add_argument("-t", action="store", type=float, default=2.0, help="time between polling")
    parser.add_argument("--temperature", action="store_true", default=False)
    parser.add_argument("--pressure", action="store_true", default=False)
    parser.add_argument("--humidity", action="store_true", default=False)
    parser.add_argument("--gas", action="store_true", default=False)
    parser.add_argument("--color", action="store_true", default=False)
    parser.add_argument("--keypress", action="store_true", default=False)
    parser.add_argument("--tap", action="store_true", default=False)
    parser.add_argument("--orientation", action="store_true", default=False)
    parser.add_argument("--quaternion", action="store_true", default=False)
    parser.add_argument("--stepcnt", action="store_true", default=False)
    parser.add_argument("--rawdata", action="store_true", default=False)
    parser.add_argument("--euler", action="store_true", default=False)
    parser.add_argument("--rotation", action="store_true", default=False)
    parser.add_argument("--heading", action="store_true", default=False)
    parser.add_argument("--gravity", action="store_true", default=False)
    parser.add_argument("--battery", action="store_true", default=False)
    parser.add_argument("--speaker", action="store_true", default=False)
    parser.add_argument("--microphone", action="store_true", default=False)
    args = parser.parse_args()

    print(f"Connecting to {args.mac_address}")
    thingy = Thingy52(args.mac_address)
    print("Connected...")
    thingy.setDelegate(MyDelegate())

    try:
        # Set LED so that we know we are connected
        thingy.ui.enable()
        thingy.ui.set_led_mode_breathe(0x01, 50, 100)  # 0x01 = RED
        print("LED set to breathe mode...")

        # Enabling selected sensors
        print("Enabling selected sensors...")
        # Environment Service
        if args.temperature:
            thingy.environment.enable()
            thingy.environment.configure(temp_int=1000)
            thingy.environment.set_temperature_notification(True)
        if args.pressure:
            thingy.environment.enable()
            thingy.environment.configure(press_int=1000)
            thingy.environment.set_pressure_notification(True)
        if args.humidity:
            thingy.environment.enable()
            thingy.environment.configure(humid_int=1000)
            thingy.environment.set_humidity_notification(True)
        if args.gas:
            thingy.environment.enable()
            thingy.environment.configure(gas_mode_int=1)
            thingy.environment.set_gas_notification(True)
        if args.color:
            thingy.environment.enable()
            thingy.environment.configure(color_int=1000)
            thingy.environment.configure(color_sens_calib=[0, 0, 0])
            thingy.environment.set_color_notification(True)
        # User Interface Service
        if args.keypress:
            thingy.ui.enable()
            thingy.ui.set_btn_notification(True)
        if args.battery:
            thingy.battery.enable()
        # Motion Service
        if args.tap:
            thingy.motion.enable()
            thingy.motion.configure(motion_freq=200)
            thingy.motion.set_tap_notification(True)
        if args.orientation:
            thingy.motion.enable()
            thingy.motion.set_orient_notification(True)
        if args.quaternion:
            thingy.motion.enable()
            thingy.motion.set_quaternion_notification(True)
        if args.stepcnt:
            thingy.motion.enable()
            thingy.motion.configure(step_int=100)
            thingy.motion.set_stepcnt_notification(True)
        if args.rawdata:
            thingy.motion.enable()
            thingy.motion.set_rawdata_notification(True)
        if args.euler:
            thingy.motion.enable()
            thingy.motion.set_euler_notification(True)
        if args.rotation:
            thingy.motion.enable()
            thingy.motion.set_rotation_notification(True)
        if args.heading:
            thingy.motion.enable()
            thingy.motion.set_heading_notification(True)
        if args.gravity:
            thingy.motion.enable()
            thingy.motion.set_gravity_notification(True)
        # Sound Service
        if args.speaker:
            thingy.sound.enable()
            thingy.sound.configure(speaker_mode=0x03)
            thingy.sound.set_speaker_status_notification(True)
            # Test speaker
            thingy.sound.play_speaker_sample(1)
        if args.microphone:
            thingy.sound.enable()
            thingy.sound.configure(microphone_mode=0x01)
            thingy.sound.set_microphone_notification(True)

        # Allow sensors time to start up (might need more time for some sensors to be ready)
        print("All requested sensors and notifications are enabled...")
        time.sleep(1.0)

        counter = 1
        while True:
            if args.battery:
                print("Battery: ", thingy.battery.read())

            if counter >= args.count:
                break

            counter += 1
            thingy.waitForNotifications(args.t)

    finally:
        thingy.disconnect()
        del thingy


if __name__ == "__main__":
    main()
